using System.Diagnostics;
using System.Data;
using System.Collections;
using Microsoft.VisualBasic;
using System.Collections.Generic;
using System;

namespace Optim_Newton1
{
	public delegate double MyFxDelegate(int nNumVars, ref double[] fX, ref double[] fParam);
	public delegate string SayFxDelegate();
	
	public class COptimNewton1
	{
		
		MyFxDelegate m_MyFx;
		
		protected double MyFxEx(int nNumVars, ref double[] fX, ref double[] fParam, ref double[] fDeltaX, double fLambda)
		{
			int i;
			double[] fXX = new double[nNumVars];
			
			for (i = 0; i < nNumVars; i++)
			{
				fXX[i] = fX[i] + fLambda * fDeltaX[i];
			}

            return m_MyFx(nNumVars, ref fXX, ref fParam);
		}
		
		protected double FirstDeriv(int nNumVars, ref double[] fX, ref double[] fParam, int nIdxI)
		{
			double fXt, h, Fp, Fm;
			
			fXt = fX[nIdxI];
			h = 0.01 *(1 + Math.Abs(fXt));
			fX[nIdxI] = fXt + h;
            Fp = m_MyFx(nNumVars, ref fX, ref fParam);
			fX[nIdxI] = fXt - h;
            Fm = m_MyFx(nNumVars, ref fX, ref fParam);
			fX[nIdxI] = fXt;
			return (Fp - Fm) / 2 / h;
		}
		
		protected double SecondDeriv(int nNumVars, ref double[] fX, ref double[] fParam, int nIdxI, int nIdxJ)
		{
			double fXt, fYt, fHX, fHY, F0, Fp;
			double Fm, Fpp, Fmm, Fpm, Fmp, fResult;
			
			// calculate second derivative?
			if (nIdxI == nIdxJ)
			{
                F0 = m_MyFx(nNumVars, ref fX, ref fParam);
				fXt = fX[nIdxI];
				fHX = 0.01 *(1 + Math.Abs(fXt));
				fX[nIdxI] = fXt + fHX;
                Fp = m_MyFx(nNumVars, ref fX, ref fParam);
				fX[nIdxI] = fXt - fHX;
                Fm = m_MyFx(nNumVars, ref fX, ref fParam);
				fX[nIdxI] = fXt;
				fResult = (Fp - 2 * F0 + Fm) /Math.Pow(fHX, 2);
			}
			else
			{
				fXt = fX[nIdxI];
				fYt = fX[nIdxJ];
				fHX = 0.01 *(1 + Math.Abs(fXt));
				fHY = 0.01 *(1 + Math.Abs(fYt));
				// calculate Fpp
				fX[nIdxI] = fXt + fHX;
				fX[nIdxJ] = fYt + fHY;
                Fpp = m_MyFx(nNumVars, ref fX, ref fParam);
				// calculate Fmm
				fX[nIdxI] = fXt - fHX;
				fX[nIdxJ] = fYt - fHY;
                Fmm = m_MyFx(nNumVars, ref fX, ref fParam);
				// calculate Fpm
				fX[nIdxI] = fXt + fHX;
				fX[nIdxJ] = fYt - fHY;
                Fpm = m_MyFx(nNumVars, ref fX, ref fParam);
				// calculate Fmp
				fX[nIdxI] = fXt - fHX;
				fX[nIdxJ] = fYt + fHY;
                Fmp = m_MyFx(nNumVars, ref fX, ref fParam);
				fX[nIdxI] = fXt;
				fX[nIdxJ] = fYt;
				fResult = (Fpp - Fmp - Fpm + Fmm) /(4 * fHX * fHY);
			}
			
			return fResult;
		}
		
		
		protected void GetFirstDerives(int nNumVars, ref double[] fX, ref double[] fParam, ref double[] fFirstDerivX)
		{
			
			int i;
			
			for (i = 0; i < nNumVars; i++)
			{
                fFirstDerivX[i] = FirstDeriv(nNumVars, ref fX, ref fParam, i);
			}
		}
		
		protected void GetSecondDerives(int nNumVars, ref double[] fX, ref double[] fParam, ref double[,] fSecondDerivX)
		{

            int i, j;
			
			for (i = 0; i < nNumVars; i++)
			{
				for (j = 0; j < nNumVars; j++)
				{
                    fSecondDerivX[i, j] = SecondDeriv(nNumVars, ref fX, ref fParam, i, j);
				}
			}
		}
		
		public double CalcOptim(int nNumVars, ref double[] fX, ref double[] fParam, ref double[] fToler, double fEpsFx, int nMaxIter, ref int nIter, ref string sErrorMsg, MyFxDelegate MyFx)
		{
			
			int i;
			double fNorm;
			double[] g = new double[nNumVars + 1];
			int[] Index = new int[nNumVars + 1];
			double[] DeltaX = new double[nNumVars + 1];
			double[,] j = new double[nNumVars + 1, nNumVars + 1];
			bool bStop;
			
			m_MyFx = MyFx;
			
			nIter = 1;
			do
			{
				nIter++;
				if (nIter > nMaxIter)
				{
					sErrorMsg = "Reached maximum iterations limit";
					break;
				}

                GetFirstDerives(nNumVars, ref fX, ref fParam, ref g);
				
				// test if gradient is shallow enough
                fNorm = MatrixLibVb.VectNorm(ref g, 2);
				if (fNorm < fEpsFx)
				{
					break;
				}

                GetSecondDerives(nNumVars, ref fX, ref fParam, ref j);
                MatrixLibVb.MV_LUDecomp(ref j, ref Index, nNumVars);
                MatrixLibVb.MV_LUBackSubst(ref j, ref Index, nNumVars, ref g);
				for (i = 0; i < nNumVars; i++)
				{
					DeltaX[i] = g[i];
					fX[i] -= DeltaX[i];
				}
				
				bStop = true;
				for (i = 0; i < nNumVars; i++)
				{
					if (Math.Abs(DeltaX[i]) > fToler[i])
					{
						bStop = false;
						break;
					}
				}
				
			} while (!bStop);

            return MyFx(nNumVars, ref fX, ref fParam);
		}
	}
	
}
